@@ -2,19 +2,25 @@ import mne
 import tensorflow as tf
 from tensorflow import keras
 from tensorflow.keras import layers
+from tensorflow.keras.layers import (LSTM, GRU, Dense, Flatten, Bidirectional, 
+    TimeDistributed, InputLayer, Activation, Reshape, concatenate, Concatenate, 
+    Dropout)
 from tensorflow.keras import backend as K
 from keras.layers.core import Lambda
+from scipy.optimize import minimize_scalar
+import datetime
 # from sklearn import linear_model
 import numpy as np
 from scipy.stats import pearsonr
 from copy import deepcopy
 from time import time
+
 from . import util
 from . import losses
+from .custom_layers import BahdanauAttention, Attention
 
-class Net(keras.Sequential):
+class Net:
     ''' The neural network class that creates and trains the model. 
-    Inherits the keras.Sequential class
     
     Attributes
     ----------
@@ -26,6 +32,8 @@ class Net(keras.Sequential):
         Number of neurons per hidden layer.
     activation_function : str
         The activation function used for each fully connected layer.
+    n_jobs : int
+        Number of jobs/ cores to use during parallel processing
 
     Methods
     -------
@@ -35,20 +43,22 @@ class Net(keras.Sequential):
     evaluate : evaluate the performance of the model
     '''
     
-    def __init__(self, fwd, n_layers=1, n_neurons=128, n_lstm_units=100, 
-        activation_function='swish', n_jobs=-1, verbose=True):
+    def __init__(self, fwd, n_dense_layers=1, n_lstm_layers=1, 
+        n_dense_units=100, n_lstm_units=100, activation_function='swish', 
+        n_jobs=-1, verbose=True):
 
-        super().__init__()
         self._embed_fwd(fwd)
         
-        self.n_layers = n_layers
-        self.n_neurons = n_neurons
+        self.n_dense_layers = n_dense_layers
+        self.n_lstm_layers = n_lstm_layers
+        self.n_dense_units = n_dense_units
         self.n_lstm_units = n_lstm_units
         self.activation_function = activation_function
         # self.default_loss = tf.keras.losses.Huber(delta=delta)
         self.default_loss = losses.weighted_huber_loss
         # self.parallel = parallel
         self.n_jobs = n_jobs
+        self.compiled = False
         self.verbose = verbose
 
     def _embed_fwd(self, fwd):
@@ -102,9 +112,10 @@ class Net(keras.Sequential):
         return eeg, sources
 
     def fit(self, *args, optimizer=None, learning_rate=0.001, 
-        validation_split=0.1, epochs=50, metrics=None, device=None, false_positive_penalty=2, 
-        delta=1., batch_size=128, loss=None, sample_weight=None, return_history=False,
-        dropout=0.2):
+        validation_split=0.1, epochs=50, metrics=None, device=None, 
+        false_positive_penalty=2, delta=1., batch_size=128, loss=None, 
+        sample_weight=None, return_history=False, dropout=0.2, patience=7, 
+        tensorboard=False):
         ''' Train the neural network using training data (eeg) and labels (sources).
         
         Parameters
@@ -156,12 +167,10 @@ class Net(keras.Sequential):
 
         self.dropout = dropout
         eeg, sources = self._handle_data_input(args)
-        self.subject = sources.subject if type(sources) == mne.SourceEstimate else sources[0].subject
-
-        if type(sources) == list:
-            self.temporal = True
-        else:
-            self.temporal = False
+        self.subject = sources.subject if type(sources) == mne.SourceEstimate \
+            else sources[0].subject
+        
+        
         # Ensure that the forward model has the same 
         # channels as the eeg object
         self._check_model(eeg)
@@ -191,50 +200,60 @@ class Net(keras.Sequential):
 
         # Early stopping
         es = tf.keras.callbacks.EarlyStopping(monitor='val_loss', \
-            mode='min', verbose=self.verbose, patience=7, restore_best_weights=True)
-            
+            mode='min', verbose=self.verbose, patience=patience, restore_best_weights=True)
+        if tensorboard:
+            log_dir = "logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
+            tensorboard_callback = tf.keras.callbacks.TensorBoard(
+                log_dir=log_dir, histogram_freq=1)
+            callbacks = [es, tensorboard_callback]
+        else:
+            callbacks = [es]
         if optimizer is None:
             optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
+            # optimizer = 'adam'
         if loss is None:
             loss = self.default_loss(weight=false_positive_penalty, delta=delta)
+            # loss = 'mse'
+
 
         elif type(loss) == list:
             loss = loss[0](*loss[1])
         if metrics is None:
             metrics = [self.default_loss(weight=false_positive_penalty, delta=delta)]
+            # metrics = ['mae']
+        
+        # Compile if it wasnt compiled before
+        if not self.compiled:
+            self.model.compile(optimizer, loss, metrics=metrics)
+            self.compiled = True
 
-        self.compile(optimizer, loss, metrics=metrics)
 
-        if self.temporal:
-            # LSTM net expects dimensions to be: (samples, time, channels)
-            x_scaled = np.swapaxes(x_scaled,1,2)
-            y_scaled = np.swapaxes(y_scaled,1,2)
-        else:
-            # Squeeze to remove empty time dimension
+        # Add empty time dimension
+        if len(np.squeeze(x_scaled).shape) == 2:
             x_scaled = np.squeeze(x_scaled)
+            x_scaled = x_scaled[:, :, np.newaxis]
+        if len(np.squeeze(y_scaled).shape) == 2:
             y_scaled = np.squeeze(y_scaled)
+            y_scaled = y_scaled[:, :, np.newaxis]
 
+        # Temporal nets expect dimensions to be: (samples, time, channels)
+        x_scaled = np.swapaxes(x_scaled,1,2)
+        y_scaled = np.swapaxes(y_scaled,1,2)
+        print(x_scaled.shape, y_scaled.shape)
         if device is None:
-            try:
-                history = super(Net, self).fit(x_scaled, y_scaled, epochs=epochs, batch_size=batch_size, shuffle=False, \
-                    validation_split=validation_split, verbose=self.verbose, callbacks=[es],
-                    sample_weight=sample_weight)
-            except:
-                history = super().fit(x_scaled, y_scaled, epochs=epochs, batch_size=batch_size, shuffle=False, \
-                    validation_split=validation_split, verbose=self.verbose, callbacks=[es],
-                    sample_weight=sample_weight)
+            history = self.model.fit(x_scaled, y_scaled, 
+                epochs=epochs, batch_size=batch_size, shuffle=True, 
+                validation_split=validation_split, verbose=self.verbose, 
+                callbacks=callbacks, sample_weight=sample_weight)
         else:
             with tf.device(device):
-                try:
-                    history = super(Net, self).fit(x_scaled, y_scaled, epochs=epochs, batch_size=batch_size, shuffle=False, \
-                        validation_split=validation_split, verbose = self.verbose, callbacks=[es],
-                        sample_weight=sample_weight)
-                except:
-                    history = super().fit(x_scaled, y_scaled, epochs=epochs, batch_size=batch_size, shuffle=False, \
-                        validation_split=validation_split, verbose = self.verbose, callbacks=[es],
-                        sample_weight=sample_weight)
+                history = self.model.fit(x_scaled, y_scaled, 
+                    epochs=epochs, batch_size=batch_size, shuffle=True, 
+                    validation_split=validation_split, verbose=self.verbose,
+                    callbacks=callbacks, sample_weight=sample_weight)
+                
         if return_history:
-            self, history
+            return self, history
         else:
             return self
 
@@ -252,17 +271,18 @@ class Net(keras.Sequential):
         eeg : numpy.ndarray
             Scaled EEG
         '''
-
+        eeg_out = deepcopy(eeg)
         # Common average ref
         for sample in range(eeg.shape[0]):
             for time in range(eeg.shape[2]):
-                eeg[sample, :, time] -= np.mean(eeg[sample, :, time])
+                eeg_out[sample, :, time] -= np.mean(eeg_out[sample, :, time])
+                eeg_out[sample, :, time] /= eeg_out[sample, :, time].std()
         
         # Normalize
-        for sample in range(eeg.shape[0]):
-            eeg[sample] /= eeg[sample].std()
+        # for sample in range(eeg.shape[0]):
+        #     eeg[sample] /= eeg[sample].std()
 
-        return eeg
+        return eeg_out
             
 
     def scale_source(self, source):
@@ -278,10 +298,12 @@ class Net(keras.Sequential):
         source : numpy.ndarray
             Scaled sources
         '''
+        source_out = deepcopy(source)
         for sample in range(source.shape[0]):
-            source[sample] /= np.max(np.abs(source[sample]))
+            for time in range(source.shape[2]):
+                source_out[sample, :, time] /= np.max(np.abs(source_out[sample, :, time]))
 
-        return source
+        return source_out
             
 
     def predict(self, *args):
@@ -307,8 +329,6 @@ class Net(keras.Sequential):
         
         eeg, _ = self._handle_data_input(args)
         
-        
-
         if isinstance(eeg, util.EVOKED_INSTANCES):
             # Ensure there are no extra channels in our EEG
             eeg = eeg.pick_channels(self.fwd.ch_names)    
@@ -337,15 +357,15 @@ class Net(keras.Sequential):
         # eeg_prep =  self._prep_eeg(eeg)
         eeg_prep = self.scale_eeg(eeg)
         
+        # Reshape to (samples, time, channels)
+        eeg_prep = np.swapaxes(eeg_prep, 1,2)
+
         # Predicted sources all in one go
-        if self.temporal:
-            eeg_prep = np.swapaxes(eeg_prep, 1,2)
         predicted_sources = self.predict_sources(eeg_prep)       
         
         # Rescale Predicitons
         # predicted_sources_scaled = self._solve_p_wrap(predicted_sources, eeg)
         predicted_sources_scaled = self._scale_p_wrap(predicted_sources, eeg)
-
         # Convert sources (numpy.ndarrays) to mne.SourceEstimates objects
         if predicted_sources.shape[-1] == 1:
             predicted_source_estimate = [util.source_to_sourceEstimate(predicted_sources_scaled[:, :, 0], self.fwd, sfreq=sfreq, tmin=tmin, subject=self.subject)]
@@ -368,29 +388,23 @@ class Net(keras.Sequential):
             3D numpy array of EEG data (samples, channels, time)
         '''
         assert len(eeg.shape)==3, 'eeg must be a 3D numpy array of dim (samples, channels, time)'
-        if not self.temporal:
-            # Predict sources all at once
-            n_samples, n_elec, n_time = eeg.shape
-            ## swap electrode and time axis
-            eeg_tmp = np.swapaxes(eeg, 1, 2)
-            ## reshape axis
-            new_shape = (n_samples*n_time, n_elec)
-            eeg_tmp = eeg_tmp.reshape(new_shape)
-            ## predict
-            try:
-                predicted_sources = super(Net, self).predict(eeg_tmp)
-            except:
-                predicted_sources = super().predict(eeg_tmp)
+        # if not self.temporal:
+        #     # Predict sources all at once
+        #     n_samples, n_time, n_elec = eeg.shape
+        #     ## reshape axis
+        #     new_shape = (n_samples*n_time, n_elec)
+        #     eeg_tmp = eeg.reshape(new_shape)
+        #     ## predict
+            
+        #     predicted_sources = self.model.predict(eeg_tmp)
+            
+        #     ## Get to old shape
+        #     predicted_sources = predicted_sources.reshape(n_samples, n_time, self.n_dipoles)
+        # else:
+        predicted_sources = self.model.predict(eeg)
+            
+        predicted_sources = np.swapaxes(predicted_sources,1,2)
 
-            ## Get to old shape
-            predicted_sources = predicted_sources.reshape(n_samples, n_time, self.n_dipoles)
-            predicted_sources = np.swapaxes(predicted_sources, 1,2)
-        else:
-            try:
-                predicted_sources = super(Net, self).predict(eeg)
-            except:
-                predicted_sources = super().predict(eeg)
-            predicted_sources = np.swapaxes(predicted_sources,1,2)
         return predicted_sources
 
     def _scale_p_wrap(self, y_est, x_true):
@@ -399,7 +413,6 @@ class Net(keras.Sequential):
         '''
         assert len(y_est.shape) == 3, 'Sources must be 3-Dimensional'
         assert len(x_true.shape) == 3, 'EEG must be 3-Dimensional'
-
         y_est_scaled = deepcopy(y_est)
 
         for trial in range(x_true.shape[0]):
@@ -463,91 +476,109 @@ class Net(keras.Sequential):
         -------
         net = Net()
         net.fit(simulation)
-        mean_squared_errors = net.evaluate(simulation)
+        mean_squared_errors = net.evaluate_mse(simulation)
         print(mean_squared_errors.mean())
         '''
 
         eeg, sources = self._handle_data_input(args)
         
-        y_hat = self.predict(eeg).data
-        y = sources.data
+        y_hat = self.predict(eeg)
+        
+        if type(y_hat) == list:
+            y_hat = np.stack([y.data for y in y_hat], axis=0)
+        else:
+            y_hat = y_hat.data
+
+        if type(sources) == list:
+            y = np.stack([y.data for y in sources], axis=0)
+        else:
+            y = sources.data
         mean_squared_errors = np.mean((y_hat - y)**2, axis=0)
 
         return mean_squared_errors
 
+    def evaluate_nmse(self, *args):
+        ''' Evaluate the model regarding normalized mean squared error
+        
+        Parameters
+        ----------
+        *args : 
+            Can be either 
+                eeg : mne.Epochs/ numpy.ndarray
+                    The simulated EEG data
+                sources : mne.SourceEstimates/ list of mne.SourceEstimates
+                    The simulated EEG data
+                or
+                simulation : esinet.simulation.Simulation
+                    The Simulation object
+
+        Return
+        ------
+        normalized_mean_squared_errors : numpy.ndarray
+            The normalized mean squared error of each sample
+
+        Example
+        -------
+        net = Net()
+        net.fit(simulation)
+        normalized_mean_squared_errors = net.evaluate_nmse(simulation)
+        print(normalized_mean_squared_errors.mean())
+        '''
+
+        eeg, sources = self._handle_data_input(args)
+        
+        y_hat = self.predict(eeg)
+        
+        if type(y_hat) == list:
+            y_hat = np.stack([y.data for y in y_hat], axis=0)
+        else:
+            y_hat = y_hat.data
+
+        if type(sources) == list:
+            y = np.stack([y.data for y in sources], axis=0)
+        else:
+            y = sources.data
+
+        for s in range(y_hat.shape[0]):
+            for t in range(y_hat.shape[2]):
+                y_hat[s, :, t] /= np.max(np.abs(y_hat[s, :, t]))
+                y[s, :, t] /= np.max(np.abs(y[s, :, t]))
+                
+        normalized_mean_squared_errors = np.mean((y_hat - y)**2, axis=0)
+
+        return normalized_mean_squared_errors
 
     def _build_model(self):
-        ''' Build the neural network architecture using the tensorflow.keras.Sequential() API. 
-        Depending on the input data this function will either build:
+        ''' Build the neural network architecture using the 
+        tensorflow.keras.Sequential() API. Depending on the input data this 
+        function will either build:
 
         (1) A simple single hidden layer fully connected ANN for single time instance data
         (2) A LSTM network for spatio-temporal prediction
         '''
-        if self.temporal:
-            # self._build_temporal_model()
-            self._build_temporal_model_v2()
-        else:
-            self._build_perceptron_model()
+     
+        self._build_perceptron_model()
         
 
         if self.verbose:
-            self.summary()
+            self.model.summary()
     
-    def _build_temporal_model(self):
-        ''' Build the temporal artificial neural network model using LSTM layers.
-        '''
-        tf.keras.backend.set_image_data_format('channels_last')
-        input_shape = (self.n_timepoints, self.n_channels)
-        self.add(layers.InputLayer(input_shape=input_shape))
-        
-        for _ in range(self.n_layers):
-            self.add(layers.Bidirectional(layers.LSTM(self.n_lstm_units, return_sequences=True, 
-                input_shape=(self.n_timepoints, self.n_channels), 
-                dropout=self.dropout, activation=self.activation_function)))
-        self.add(layers.Flatten())
-        self.add(layers.Dense(int(self.n_timepoints*self.n_dipoles), 
-            activation='linear'))
-        self.add(layers.Reshape((self.n_timepoints, self.n_dipoles)))
-        self.add(layers.Activation('linear'))
-        
-        self.build(input_shape=input_shape)
-        
-    def _build_temporal_model_v2(self):
-        ''' Build the temporal artificial neural network model using LSTM layers.
-        '''
-
-        tf.keras.backend.set_image_data_format('channels_last')
-        input_shape = (None, self.n_channels)
-        self.add(layers.InputLayer(input_shape=input_shape))
-        
-        # LSTM layers
-        for _ in range(self.n_layers):
-            self.add(layers.LSTM(self.n_lstm_units, return_sequences=True, 
-                input_shape=input_shape, 
-                dropout=self.dropout, activation=self.activation_function))
-
-        # For-each layer:
-        self.add(layers.TimeDistributed(
-            layers.Dense(self.n_dipoles, activation='linear'))
-        )
-
-        self.build(input_shape=input_shape)
-
     def _build_perceptron_model(self):
         ''' Build the artificial neural network model using Dense layers.
         '''
+        input_shape = (None, None, self.n_channels)
+        tf.keras.backend.set_image_data_format('channels_last')
+
+        self.model = keras.Sequential()
         # Add hidden layers
-        for i in range(self.n_layers):
-            self.add(layers.Dense(units=self.n_neurons,
-                                activation=self.activation_function))
+        for _ in range(self.n_dense_layers):
+            self.model.add(TimeDistributed(Dense(units=self.n_dense_units,
+                                activation=self.activation_function)))
         # Add output layer
-        self.add(layers.Dense(self.n_dipoles, 
-            activation='linear'))
+        self.model.add(TimeDistributed(Dense(self.n_dipoles, activation='linear')))
         
         # Build model with input layer
-        self.build(input_shape=(None, self.n_channels))
-
-    
+        self.model.build(input_shape=input_shape)
 
 
     def _check_model(self, eeg):
@@ -561,7 +592,7 @@ class Net(keras.Sequential):
 
         '''
         # Dont do anything if model is already built.
-        if self.built:
+        if self.compiled:
             return
         
         # Else assure that channels are appropriate
@@ -597,51 +628,52 @@ class Net(keras.Sequential):
         x_true = np.squeeze(np.array(x_true))
         # Get EEG from predicted source using leadfield
         x_est = np.matmul(self.leadfield, y_est)
+
         gfp_true = np.std(x_true)
         gfp_est = np.std(x_est)
         scaler = gfp_true / gfp_est
         y_est_scaled = y_est * scaler
         return y_est_scaled
         
-    # def solve_p(self, y_est, x_true):
-    #     '''
-    #     Parameters
-    #     ---------
-    #     y_est : numpy.ndarray
-    #         The estimated source vector.
-    #     x_true : numpy.ndarray
-    #         The original input EEG vector.
+    def solve_p(self, y_est, x_true):
+        '''
+        Parameters
+        ---------
+        y_est : numpy.ndarray
+            The estimated source vector.
+        x_true : numpy.ndarray
+            The original input EEG vector.
         
-    #     Return
-    #     ------
-    #     y_scaled : numpy.ndarray
-    #         The scaled estimated source vector.
+        Return
+        ------
+        y_scaled : numpy.ndarray
+            The scaled estimated source vector.
         
-    #     '''
-    #     # Check if y_est is just zeros:
-    #     if np.max(y_est) == 0:
-    #         return y_est
-    #     y_est = np.squeeze(np.array(y_est))
-    #     x_true = np.squeeze(np.array(x_true))
-    #     # Get EEG from predicted source using leadfield
-    #     x_est = np.matmul(self.leadfield, y_est)
+        '''
+        # Check if y_est is just zeros:
+        if np.max(y_est) == 0:
+            return y_est
+        y_est = np.squeeze(np.array(y_est))
+        x_true = np.squeeze(np.array(x_true))
+        # Get EEG from predicted source using leadfield
+        x_est = np.matmul(self.leadfield, y_est)
 
-    #     # optimize forward solution
-    #     tol = 1e-3
-    #     options = dict(maxiter=1000, disp=False)
+        # optimize forward solution
+        tol = 1e-3
+        options = dict(maxiter=1000, disp=False)
 
-    #     # base scaling
-    #     rms_est = np.mean(np.abs(x_est))
-    #     rms_true = np.mean(np.abs(x_true))
-    #     base_scaler = rms_true / rms_est
+        # base scaling
+        rms_est = np.mean(np.abs(x_est))
+        rms_true = np.mean(np.abs(x_true))
+        base_scaler = rms_true / rms_est
 
         
-    #     opt = minimize_scalar(self.correlation_criterion, args=(self.leadfield, y_est* base_scaler, x_true), \
-    #         bounds=(0, 1), method='bounded', options=options, tol=tol)
+        opt = minimize_scalar(self.correlation_criterion, args=(self.leadfield, y_est* base_scaler, x_true), \
+            bounds=(0, 1), method='bounded', options=options, tol=tol)
         
-    #     scaler = opt.x
-    #     y_scaled = y_est * scaler * base_scaler
-    #     return y_scaled
+        scaler = opt.x
+        y_scaled = y_est * scaler * base_scaler
+        return y_scaled
 
     @staticmethod
     def correlation_criterion(scaler, leadfield, y_est, x_true):
