@@ -5,6 +5,7 @@ import random
 from joblib import Parallel, delayed
 from tqdm.notebook import tqdm
 import colorednoise as cn
+import mne
 from time import time
 from . import util
 
@@ -367,6 +368,12 @@ class Simulation:
         # since they can have entirely different scales.
         coil_types = [ch['coil_type'] for ch in self.info['chs']]
         coil_types_set = list(set(coil_types))
+        if len(coil_types_set)>1:
+            msg = f'Simulations attempted with more than one channel type \
+                ({coil_types_set}) may result in unexpected behavior. Please \
+                select one channel type in your data only'
+            raise ValueError(msg)
+            
         coil_types_set = np.array([int(i) for i in coil_types_set])
         
         coil_type_assignments = np.array(
@@ -410,12 +417,18 @@ class Simulation:
         short_shape = (sources_tmp.shape[0], 
             sources_tmp.shape[1]*sources_tmp.shape[2])
         sources_tmp = sources_tmp.reshape(short_shape)
+        # Scale to allow for lower precision
+        scaler = 1/sources_tmp.max()
+        sources_tmp *= scaler
         # Perform Matmul
-        result = np.matmul(leadfield, sources_tmp)
+        result = np.matmul(
+            leadfield.astype(np.float32), sources_tmp.astype(np.float32))
         # Reshape result
         result = result.reshape(result.shape[0], n_samples, n_timepoints)
         # swap axes to correct order
         result = np.swapaxes(result,0,1)
+        # Rescale
+        result /= scaler
         return result
 
 
@@ -532,3 +545,51 @@ class Simulation:
         with open(file_name, 'wb') as f:
             pkl.dump(self, f)
 
+    def to_nontemporal(self):
+        ''' Converts the internal data representation from temporal to 
+        non-temporal. 
+        
+        Specifically, this changes the shape of sources from a
+        list of mne.sourceEstimate to a single mne.sourceEstimate in which the 
+        time dimension holds a concatenation of timepoints and samples.
+
+        The eeg data is reshaped from (samples, channels, time points) to 
+        (samples*time points, channels, 1).
+
+        Parameters
+        ----------
+        
+
+        Return
+        ------
+        self : esinet.Simulation
+            Method returns itself for convenience
+
+        '''
+        if not self.temporal:
+            print('This Simulation() instance is already non-temporal')
+            return self
+
+        self.temporal = False
+        self.settings['duration_of_trial'] = 0
+
+        eeg_data_lstm = self.eeg_data.get_data()
+        # Reshape EEG data
+        eeg_data_single = np.expand_dims(np.vstack(np.swapaxes(eeg_data_lstm, 1,2)), axis=-1)
+        # Pack into mne.EpochsArray object
+        epochs_single = mne.EpochsArray(eeg_data_single, self.eeg_data.info, 
+            tmin=self.eeg_data.tmin, verbose=0)
+        # Store the newly shaped data
+        self.eeg_data = epochs_single
+        
+        # Reshape Source data
+        source_data = np.vstack(np.swapaxes(np.stack(
+            [source.data for source in self.source_data], axis=0), 1,2)).T
+        # Pack into mne.SourceEstimate object
+        source_single = deepcopy(self.source_data[0])
+        source_single.data = source_data
+        self.source_data = source_single
+        
+        return self
+        
+        
\ No newline at end of file
